@@ -1,5 +1,5 @@
 /*
-% Copyright (C) 2003 - 2011 GraphicsMagick Group
+% Copyright (C) 2003 - 2016 GraphicsMagick Group
 % Copyright (C) 2002 ImageMagick Studio
 % Copyright 1991-1999 E. I. du Pont de Nemours and Company
 %
@@ -249,6 +249,16 @@ static int SGIDecode(const unsigned long bytes_per_pixel,
   return 0;
 }
 
+#define ThrowSGIReaderException(code_,reason_,image_) \
+{ \
+  MagickFreeMemory(iris_pixels) \
+  MagickFreeMemory(max_packets); \
+  MagickFreeMemory(offsets); \
+  MagickFreeMemory(runlength); \
+  MagickFreeMemory(scanline) \
+  ThrowReaderException(code_,reason_,image_); \
+}
+
 static Image *ReadSGIImage(const ImageInfo *image_info,ExceptionInfo *exception)
 {
   Image
@@ -274,15 +284,20 @@ static Image *ReadSGIImage(const ImageInfo *image_info,ExceptionInfo *exception)
   SGIInfo
     iris_info;
 
+  unsigned long
+    *offsets = (unsigned long *) NULL,
+    *runlength = (unsigned long *) NULL;
+
   unsigned char
-    *iris_pixels;
+    *iris_pixels = (unsigned char *) NULL,
+    *max_packets = (unsigned char *) NULL,
+    *scanline = (unsigned char *) NULL;
 
   unsigned int
     status;
 
-  unsigned long
-    bytes_per_pixel,
-    number_pixels;
+  size_t
+    bytes_per_pixel;
 
   /*
     Open image file.
@@ -325,7 +340,9 @@ static Image *ReadSGIImage(const ImageInfo *image_info,ExceptionInfo *exception)
 		      iris_info.dummy2);
 
       (void) LogMagickEvent(CoderEvent,GetMagickModule(),
-			    "    Header: Storage=%u, BPC=%u, Dimension=%u, XSize=%u, YSize=%u, ZSize=%u, PixMin=%u, PixMax=%u, image_name=\"%.79s\", color_map=%u",
+			    "    Header: Storage=%u, BPC=%u, Dimension=%u, "
+                            "XSize=%u, YSize=%u, ZSize=%u, PixMin=%u, "
+                            "PixMax=%u, image_name=\"%.79s\", color_map=%u",
 			    (unsigned int) iris_info.storage,
 			    (unsigned int) iris_info.bytes_per_pixel,
 			    (unsigned int) iris_info.dimension,
@@ -378,6 +395,10 @@ static Image *ReadSGIImage(const ImageInfo *image_info,ExceptionInfo *exception)
 	  ThrowReaderException(CorruptImageError,ImproperImageHeader,image);
 	}
 
+      /* We only support zsize up to 4. Code further on assumes that. */
+      if (iris_info.zsize > 4)
+        ThrowReaderException(CorruptImageError,ImproperImageHeader,image);
+
       if (iris_info.dimension == 1)
 	{
 	  /*
@@ -467,30 +488,28 @@ static Image *ReadSGIImage(const ImageInfo *image_info,ExceptionInfo *exception)
 	if (image->scene >= (image_info->subimage+image_info->subrange-1))
 	  break;
 
+      if (CheckImagePixelLimits(image, exception) != MagickPass)
+        ThrowReaderException(ResourceLimitError,ImagePixelLimitExceeded,image);
+
       /*
 	Allocate SGI pixels.
       */
       bytes_per_pixel=iris_info.bytes_per_pixel;
-      number_pixels=iris_info.xsize*iris_info.ysize;
-      if (4*bytes_per_pixel*number_pixels < number_pixels) /* Overflow? */
-	ThrowReaderException(CorruptImageError,InsufficientImageDataInFile,image);
-      iris_pixels=MagickAllocateMemory(unsigned char *,
-				       4*bytes_per_pixel*number_pixels);
+      iris_pixels=MagickAllocateArray(unsigned char *,
+                                      MagickArraySize(4,bytes_per_pixel),
+                                      MagickArraySize(iris_info.xsize,iris_info.ysize));
       if (iris_pixels == (unsigned char *) NULL)
-	ThrowReaderException(ResourceLimitError,MemoryAllocationFailed,image);
+	ThrowSGIReaderException(ResourceLimitError,MemoryAllocationFailed,image);
       if (iris_info.storage != 0x01)
 	{
-	  unsigned char
-	    *scanline;
-
 	  /*
 	    Read standard image format.
 	  */
-	  scanline=MagickAllocateMemory(unsigned char *,
-					bytes_per_pixel*iris_info.xsize);
+	  scanline=MagickAllocateArray(unsigned char *,
+                                       bytes_per_pixel,iris_info.xsize);
 	  if (scanline == (unsigned char *) NULL)
-	    ThrowReaderException(ResourceLimitError,MemoryAllocationFailed,
-				 image);
+	    ThrowSGIReaderException(ResourceLimitError,MemoryAllocationFailed,
+                                    image);
 	  (void) LogMagickEvent(CoderEvent,GetMagickModule(),
 				"   Reading SGI scanlines"); 
 	  for (z=0; z < (int) iris_info.zsize; z++)
@@ -502,8 +521,8 @@ static Image *ReadSGIImage(const ImageInfo *image_info,ExceptionInfo *exception)
 				  (char *) scanline);
 		  if (EOFBlob(image))
 		    {
-		      ThrowReaderException(CorruptImageError,
-					   UnexpectedEndOfFile, image);
+		      ThrowSGIReaderException(CorruptImageError,
+                                              UnexpectedEndOfFile, image);
 		      break;
 		    }
 		  if (bytes_per_pixel == 2)
@@ -527,39 +546,36 @@ static Image *ReadSGIImage(const ImageInfo *image_info,ExceptionInfo *exception)
 	}
       else
 	{
-	  unsigned char
-	    *max_packets;
-
 	  unsigned int
 	    data_order;
 
 	  unsigned long
-	    offset,
-	    *offsets,
-	    *runlength;
+	    offset;
 
 	  /*
 	    Read runlength-encoded image format.
 	  */
-	  offsets=MagickAllocateMemory(unsigned long *,iris_info.ysize*
-				       iris_info.zsize*sizeof(unsigned long));
+	  offsets=MagickAllocateArray(unsigned long *,iris_info.ysize,
+                                      MagickArraySize(iris_info.zsize,
+                                                      sizeof(unsigned long)));
 	  max_packets=MagickAllocateArray(unsigned char *,iris_info.xsize+10,4);
-	  runlength=MagickAllocateMemory(unsigned long *,iris_info.ysize*
-					 iris_info.zsize*sizeof(unsigned long));
+	  runlength=MagickAllocateArray(unsigned long *,iris_info.ysize,
+                                        MagickArraySize(iris_info.zsize,
+                                                        sizeof(unsigned long)));
 	  if ((offsets == (unsigned long *) NULL) ||
 	      (max_packets == (unsigned char *) NULL) ||
 	      (runlength == (unsigned long *) NULL))
-	    ThrowReaderException(ResourceLimitError,MemoryAllocationFailed,
-				 image);
+            ThrowSGIReaderException(ResourceLimitError,MemoryAllocationFailed,
+                                    image);
 	  for (i=0; i < (int) (iris_info.ysize*iris_info.zsize); i++)
 	    offsets[i]=ReadBlobMSBLong(image);
 	  for (i=0; i < (int) (iris_info.ysize*iris_info.zsize); i++)
 	    {
 	      runlength[i]=ReadBlobMSBLong(image);
 	      if (EOFBlob(image))
-		ThrowReaderException(CorruptImageError,UnexpectedEndOfFile,image);
+		ThrowSGIReaderException(CorruptImageError,UnexpectedEndOfFile,image);
 	      if (runlength[i] > ((unsigned long) 4*iris_info.xsize+10))
-		ThrowReaderException(CorruptImageError,UnableToRunlengthDecodeImage,image);
+		ThrowSGIReaderException(CorruptImageError,UnableToRunlengthDecodeImage,image);
 	    }
 	  /*
 	    Check data order.
@@ -590,16 +606,16 @@ static Image *ReadSGIImage(const ImageInfo *image_info,ExceptionInfo *exception)
 				      (char *) max_packets);
 		      if (EOFBlob(image))
 			{
-			  ThrowReaderException(CorruptImageError,
-					       UnexpectedEndOfFile, image);
+			  ThrowSGIReaderException(CorruptImageError,
+                                                  UnexpectedEndOfFile, image);
 			  break;
 			}
 		      offset+=runlength[y+z*iris_info.ysize];
 		      if (SGIDecode(bytes_per_pixel,max_packets,p+bytes_per_pixel*z,
 				    runlength[y+z*iris_info.ysize]/bytes_per_pixel,
 				    iris_info.xsize) == -1)
-			ThrowReaderException(CorruptImageError,
-					     UnableToRunlengthDecodeImage,image); 
+			ThrowSGIReaderException(CorruptImageError,
+                                                UnableToRunlengthDecodeImage,image); 
 		      p+=(iris_info.xsize*4*bytes_per_pixel);
 		    }
 		}
@@ -620,16 +636,16 @@ static Image *ReadSGIImage(const ImageInfo *image_info,ExceptionInfo *exception)
 				      (char *) max_packets);
 		      if (EOFBlob(image))
 			{
-			  ThrowReaderException(CorruptImageError,
-					       UnexpectedEndOfFile, image);
+			  ThrowSGIReaderException(CorruptImageError,
+                                                  UnexpectedEndOfFile, image);
 			  break;
 			}
 		      offset+=runlength[y+z*iris_info.ysize];
 		      if (SGIDecode(bytes_per_pixel,max_packets,p+bytes_per_pixel*z,
 				    runlength[y+z*iris_info.ysize]/bytes_per_pixel,
 				    iris_info.xsize) == -1)
-			ThrowReaderException(CorruptImageError,
-					     UnableToRunlengthDecodeImage,image); 
+			ThrowSGIReaderException(CorruptImageError,
+                                                UnableToRunlengthDecodeImage,image); 
 		    }
 		  p+=(iris_info.xsize*4*bytes_per_pixel);
 		  if (EOFBlob(image))
@@ -713,8 +729,8 @@ static Image *ReadSGIImage(const ImageInfo *image_info,ExceptionInfo *exception)
 	    Create grayscale map.
 	  */
 	  if (!AllocateImageColormap(image,image->colors))
-	    ThrowReaderException(ResourceLimitError,MemoryAllocationFailed,
-				 image);
+	    ThrowSGIReaderException(ResourceLimitError,MemoryAllocationFailed,
+                                    image);
 	  /*
 	    Convert SGI image to PseudoClass pixel packets.
 	  */
@@ -770,8 +786,8 @@ static Image *ReadSGIImage(const ImageInfo *image_info,ExceptionInfo *exception)
       MagickFreeMemory(iris_pixels);
       if (EOFBlob(image))
 	{
-	  ThrowReaderException(CorruptImageError,UnexpectedEndOfFile,
-			       image);
+	  ThrowSGIReaderException(CorruptImageError,UnexpectedEndOfFile,
+                                  image);
 	  break;
 	}
     } while (0);
@@ -925,6 +941,16 @@ static size_t SGIEncode(unsigned char *pixels,size_t count,
   return(q-packets);
 }
 
+#define ThrowSGIWriterException(code_,reason_,image_) \
+{ \
+  MagickFreeMemory(scanline) \
+  MagickFreeMemory(offsets); \
+  MagickFreeMemory(packets); \
+  MagickFreeMemory(runlength); \
+  MagickFreeMemory(iris_pixels); \
+  ThrowWriterException(code_,reason_,image_); \
+}
+
 static unsigned int WriteSGIImage(const ImageInfo *image_info,Image *image)
 {
   long
@@ -944,9 +970,14 @@ static unsigned int WriteSGIImage(const ImageInfo *image_info,Image *image)
   register unsigned char
     *q;
 
+  unsigned long
+    *offsets = (unsigned long *) NULL,
+    *runlength = (unsigned long *) NULL;
+
   unsigned char
-    *iris_pixels,
-    *packets;
+    *iris_pixels = (unsigned char *) NULL,
+    *packets = (unsigned char *) NULL,
+    *scanline = (unsigned char *) NULL;
 
   unsigned int
     status;
@@ -1046,7 +1077,7 @@ static unsigned int WriteSGIImage(const ImageInfo *image_info,Image *image)
       number_pixels=image->columns*image->rows;
       iris_pixels=MagickAllocateMemory(unsigned char *,4*number_pixels);
       if (iris_pixels == (unsigned char *) NULL)
-	ThrowWriterException(ResourceLimitError,MemoryAllocationFailed,image);
+	ThrowSGIWriterException(ResourceLimitError,MemoryAllocationFailed,image);
       /*
 	Convert image pixels to uncompressed SGI pixels.
       */
@@ -1072,16 +1103,13 @@ static unsigned int WriteSGIImage(const ImageInfo *image_info,Image *image)
 	}
       if (image_info->compression == NoCompression)
 	{
-	  unsigned char
-	    *scanline;
-
 	  /*
 	    Write uncompressed SGI pixels.
 	  */
 	  scanline=MagickAllocateMemory(unsigned char *,iris_info.xsize);
 	  if (scanline == (unsigned char *) NULL)
-	    ThrowWriterException(ResourceLimitError,MemoryAllocationFailed,
-				 image);
+	    ThrowSGIWriterException(ResourceLimitError,MemoryAllocationFailed,
+                                    image);
 	  for (z=0; z < (int) iris_info.zsize; z++)
 	    {
 	      q=iris_pixels+z;
@@ -1102,25 +1130,26 @@ static unsigned int WriteSGIImage(const ImageInfo *image_info,Image *image)
 	  unsigned long
 	    length,
 	    number_packets,
-	    offset,
-	    *offsets,
-	    *runlength;
+	    offset;
 
 	  /*
 	    Convert SGI uncompressed pixels.
 	  */
-	  offsets=MagickAllocateMemory(unsigned long *,iris_info.ysize*
-				       iris_info.zsize*sizeof(unsigned long));
-	  packets=MagickAllocateMemory(unsigned char *,
-				       4*(2*iris_info.xsize+10)*image->rows);
-	  runlength=MagickAllocateMemory(unsigned long *,iris_info.ysize*
-					 iris_info.zsize*sizeof(unsigned long));
+	  offsets=MagickAllocateArray(unsigned long *,iris_info.ysize,
+                                      MagickArraySize(iris_info.zsize,
+                                                      sizeof(unsigned long)));
+	  packets=MagickAllocateArray(unsigned char *,
+                                      4*(2*(size_t) iris_info.xsize+10),
+                                      image->rows);
+	  runlength=MagickAllocateArray(unsigned long *,iris_info.ysize,
+                                        MagickArraySize(iris_info.zsize,
+                                                        sizeof(unsigned long)));
 	  if ((offsets == (unsigned long *) NULL) ||
 	      (packets == (unsigned char *) NULL) ||
 	      (runlength == (unsigned long *) NULL))
-	    ThrowWriterException(ResourceLimitError,MemoryAllocationFailed,
-				 image);
-	  offset=512+4*2*(iris_info.ysize*iris_info.zsize);
+            ThrowSGIWriterException(ResourceLimitError,MemoryAllocationFailed,
+                                    image);
+	  offset=512+4*2*((size_t) iris_info.ysize*iris_info.zsize);
 	  number_packets=0;
 	  q=iris_pixels;
 	  for (y=0; y < (long) iris_info.ysize; y++)
